#!/bin/bash

# spdx-license-identifier: gpl-3.0-only
# Copyright (C) 2022 Michał Góral

[[ -n "${LIBGPAK_BUILD_PACKAGES_SH:-}" ]] && return
LIBGPAK_BUILD_PACKAGES_SH=1

source "${LIBRARY}/spinner.sh"
source "${LIBRARY}/message.sh"
source "${LIBRARY}/path.sh"
source "${LIBRARY}/exit_codes.sh"

_cleanup_container() {
    local contid=$1
    docker container stop --time=0  "$contid" >/dev/null 2>&1
}

_trap_interrupted() {
    local sig=$1; shift
    local contid=$1; shift
    warning "$@"

    _cleanup_container "${contid}"

    trap - "$sig"
    kill "-$sig" "$BASHPID"
}

makedeb_local() {
    local distro="$1"
    local pkgdir="$2"
    local pkgname=$(basename -- "$pkgdir")

    (
        export BUILDDIR="${BUILD_DIR}/${distro}/.build/${pkgname}/build"
        export SRCDEST="${BUILD_DIR}/${distro}/.build/${pkgname}/src"
        export SRCPKGDEST="${BUILD_DIR}/${distro}/.build/${pkgname}/srcpkg"
        export PKGDEST="${BUILD_DIR}/${distro}/${pkgname}"
        export GPAK_DISTRO="${distro}"

        cd "${pkgdir}"
        makedebargs=( makedeb --no-confirm --pass-env )

        if [[ -e "PKGBUILD.${distro}" ]]; then
            msg2 "Using distro-specific PKGBUILD.%s" "${distro}"
            makedebargs+=( "--file" "PKGBUILD.${distro}" )
        fi

        "${makedebargs[@]}"
    )
}

makedeb_docker() {
    local distro="$1"
    local pkgdir="$2"
    local pkgname=$(basename -- "$pkgdir")
    local container="gpak:${distro}"
    local outdir="$(make_absolute_path "${BUILD_DIR}/${distro}/${pkgname}")"

    local contname="gpak_${distro}_${pkgname}"
    local contid=$(docker container ls --all --quiet --filter="name=${contname}")

    if [[ -n "${contid}" ]] && (( ! CACHE )); then
        docker  container rm "${contid}"
        unset contid
    fi

    if [[ -z "${contid}" ]]; then
        contargs=(
            --name "${contname}"
            -v "${pkgdir}:/script"
            -v "${outdir}:/out"
            -e "BUILDDIR=/build/${pkgname}/build"
            -e "SRCDEST=/build/${pkgname}/sources"
            -e "SRCPKGDEST=/build/${pkgname}/srcpkg"
            -e "PKGDEST=/out"
            -e "GPAK_DISTRO=${distro}"
            --workdir="/script"
        )
        (( ! CACHE )) && contargs+=("--rm")

        makedebargs=( makedeb --sync-deps --no-color --no-confirm --pass-env )
        if [[ -e "${pkgdir}/PKGBUILD.${distro}" ]]; then
            msg2 "Using distro-specific PKGBUILD.%s" "${distro}"
            makedebargs+=( "--file" "PKGBUILD.${distro}" )
        fi

        contargs+=(
            "${container}"
            "${makedebargs[@]}"
        )

        contid=$(docker container create "${contargs[@]}")
    fi

    mkdir -p "${outdir}"

    trap "_trap_interrupted INT '${contid}' Interrupted" INT  # ignored when in a subshell
    trap "_trap_interrupted TERM '${contid}' Terminated" TERM

    # We don't know what programs run inside the docker. Maybe they are signal
    # eaters? How to handle them?
    #
    # The trick is to run docker asynchronously and wait for it. In this case
    # control returns to Bash which can correctly handle signals and traps.
    #
    # ref: https://www.gnu.org/software/bash/manual/bash.html#Signals
    docker start -a "${contid}" &
    wait $!
    ret=$?

    trap - INT TERM
    _cleanup_container "${contid}"
    return $ret
}

build_package() {
    local distro="$1" ; shift
    local pkg="$1" ; shift
    local ret=0

    if [[ -e "${pkg}/PKGBUILD" ]] || [[ -e "${pkg}/PKGBUILD.${distro}" ]]; then
        if (( BUILD_LOCAL )); then
            makedeb_local "$distro" "$pkg"
            [[ $? -ne 0 ]] && ret=$E_BUILD_ERROR
        else
            makedeb_docker "$distro" "$pkg"
            [[ $? -ne 0 ]] && ret=$E_BUILD_ERROR
        fi
    else
        error "Package doesn't have PKGBUILD file: %s" "$pkg"
        ret=$E_BUILD_ERROR
    fi

    return $ret
}

build_distro_packages() {
    local distro="$1" ; shift
    local packages=("$@")
    local ret=0

    (( BUILD_LOCAL )) && local type="local" || local type="docker"

    for pkg in "${packages[@]}"; do
        local pkgname=$(basename -- "${pkg}")
        spinmsg "Building %s for %s (%s)..." "${pkgname}" "${distro}" "${type}"
        spinner build_package "${distro}" "${pkg}"
        [[ $? -ne 0 ]] && ret=$E_BUILD_ERROR
    done

    return $ret
}

build_packages() {
    local req_packages=(${@/#/"${PACKAGES_DIR}/"})

    local ret=0
    for distro in "${DISTRIBUTIONS[@]}"; do
        build_distro_packages "$distro" "${req_packages[@]}"
        (( $? )) && ret=$E_BUILD_ERROR
    done
    return $ret
}
