#!/bin/bash

# spdx-license-identifier: gpl-3.0-only
# Copyright (C) 2022 Michał Góral

[[ -n "${LIBGPAK_SPINNER_SH:-}" ]] && return
LIBGPAK_SPINNER_SH=1

source "${LIBRARY}/message.sh"
source "${LIBRARY}/colors.sh"

_SPINMSG=""
_SPINNER_RUNNING=0

_erase_line() {
    printf "\r"
    linelen=$(tput cols)
    local i  # quite important, otherwise this i will overwrite i in spinner()
    for (( i=0; i<$linelen; i++ )); do
        printf " "
    done
    printf "\r"
}

_cleanup_spinner() {
    local subcommand_pid=$1
    local status=$2
    local print_nl=${3:-0}

    tput cnorm  # show cursor

    (( subcommand_pid )) && kill -TERM "$subcommand_pid" &> /dev/null

    _erase_line
    printf "[#] %s | %s" "${_SPINMSG}" "${status}"
    (( print_nl )) && printf "\n"
    _SPINMSG=""
    _SPINNER_RUNNING=0
}

_killme() {
    local sig=$1
    local subcommand_pid=$2

    # we're sending SIGTERM because Bash ignores SIGINT for subshells and
    # there's no way to re-enable it
    kill -TERM "$subcommand_pid" &> /dev/null
    wait $subcommand_pid

    trap - "$sig"
    kill "-$sig" "$BASHPID" &> /dev/null
}

_trim_line() {
    # trims (written here because bash parses spaces as file names):
    # 1. trim all gpak-like and makedeb message markers
    # 2. trim whitespaces from the left
    # 3. trim whitespaces from the right
    # 4. remove control characters
    sed -e 's/\[[#!-][^ ]\+//g' \
        -e 's/^[ \t]\+//' \
        -e 's/[ \t]\+$//' \
        -e 's/[\r\n]//g'
}

_get_last_info() {
    local maxlen=20
    s=$(tail -1 "${GPAK_LOG}" | _trim_line)
    if [[ ${#s} -gt $(( $maxlen - 3 )) ]]; then
        s=$(printf "%s..." "$(printf "%s" "$s" | cut -c -$maxlen)")
    fi
    printf "%s" "$s"
}

spinner() {
    if [[ ! -t 1 ]] || (( UGLY )); then
        msg "%s" "${_SPINMSG}"
        "$@"
        return $?
    fi

    msg "%s" "${_SPINMSG}" >> "${GPAK_LOG}"

    # Prevent nested spinners which will screw up (leak) file
    # descriptors by spawning several sub-subshells. Wanna guess
    # how I learned it?
    if (( ${_SPINNER_RUNNING} )); then
        "$@"
        return $?
    fi

    _SPINNER_RUNNING=1

    local spin=( ⣽ ⣾ ⣷ ⣯ ⣟  ⡿ ⢿ ⣻ )
    local i=0

    # Run a command in a subshell. Technically it's not necessary, but we avoid
    # screwing up global resources by commands this way (e.g. traps). Caveat:
    # we must remember to propagate signals!
    ("$@") >> ${GPAK_LOG} 2>&1 &
    local pid=$!

    trap "_killme TERM '$pid'" TERM
    trap "_killme INT '$pid'" INT
    trap "_cleanup_spinner '$pid' '${RED}ABORTED${ALL_OFF}' 0" EXIT

    local ts_start=$(date +%s)

    tput civis  # hide cursor
    while kill -0 $pid 2>/dev/null; do
        local ts_now=$(date +%s)
        local time_passed=$((ts_now - ts_start))
        local info="$(_get_last_info)"
        i=$(((i + 1) % ${#spin[@]}))
        _erase_line
        printf "%s %s | %s s | %s" "${spin[${i}]}" "${_SPINMSG}" "${time_passed}" "${info}"
        sleep 0.15
    done

    wait $pid
    ret=$?

    trap - EXIT INT TERM

    (( ret )) && local status="${RED}fail${ALL_OFF}" || local status="${GREEN}ok${ALL_OFF}"

    _cleanup_spinner 0 "${time_passed} s ${status}" 1
    return $ret
}

spinmsg() {
    local mesg=$1; shift
    _SPINMSG=$(printf "${mesg}" "$@")
}
